@@ -6,10 +6,10 @@
 /*                                                                           */
 /*                                                                           */
 /*                                                                           */
-/* (C) 1998-2003 Ullrich von Bassewitz                                       */
-/*               Rmerstrae 52                                              */
-/*               D-70794 Filderstadt                                         */
-/* EMail:        uz@cc65.org                                                 */
+/* (C) 1998-2009, Ullrich von Bassewitz                                      */
+/*                Roemerstrasse 52                                           */
+/*                D-70794 Filderstadt                                        */
+/* EMail:         uz@cc65.org                                                */
 /*                                                                           */
 /*                                                                           */
 /* This software is provided 'as-is', without any expressed or implied       */
@@ -84,7 +84,7 @@
 
 
 /* Keyword we're about to handle */
-static char Keyword [sizeof (SVal)+1];
+static StrBuf Keyword = STATIC_STRBUF_INITIALIZER;
 
 /* Segment stack */
 #define MAX_PUSHED_SEGMENTS     16
@@ -127,6 +127,11 @@ static unsigned char OptionalAddrSize (void)
     if (Tok == TOK_COLON) {
         NextTok ();
         AddrSize = ParseAddrSize ();
+        if (!ValidAddrSizeForCPU (AddrSize)) {
+            /* Print an error and reset to default */
+            Error ("Invalid address size specification for current CPU");
+            AddrSize = ADDR_SIZE_DEFAULT;
+        }
         NextTok ();
     }
     return AddrSize;
@@ -165,6 +170,33 @@ static void SetBoolOption (unsigned char* Flag)
 
 
 
+static void ExportWithAssign (SymEntry* Sym, unsigned char AddrSize, unsigned Flags)
+/* Allow to assign the value of an export in an .export statement */
+{
+    /* The name and optional address size spec may be followed by an assignment
+     * or equal token.
+     */
+    if (Tok == TOK_ASSIGN || Tok == TOK_EQ) {
+
+        /* Assignment means the symbol is a label */
+        if (Tok == TOK_ASSIGN) {
+            Flags |= SF_LABEL;
+        }
+
+        /* Skip the assignment token */
+        NextTok ();
+
+        /* Define the symbol with the expression following the '=' */
+        SymDef (Sym, Expression(), ADDR_SIZE_DEFAULT, Flags);
+
+    }
+
+    /* Now export the symbol */
+    SymExport (Sym, AddrSize, Flags);
+}
+
+
+
 static void ExportImport (void (*Func) (SymEntry*, unsigned char, unsigned),
                           unsigned char DefAddrSize, unsigned Flags)
 /* Export or import symbols */
@@ -181,7 +213,7 @@ static void ExportImport (void (*Func) (SymEntry*, unsigned char, unsigned),
      	}
 
         /* Find the symbol table entry, allocate a new one if necessary */
-        Sym = SymFind (CurrentScope, SVal, SYM_ALLOC_NEW);
+        Sym = SymFind (CurrentScope, &SVal, SYM_ALLOC_NEW);
 
         /* Skip the name */
         NextTok ();
@@ -211,7 +243,7 @@ static long IntArg (long Min, long Max)
  * and return -1 in this case.
  */
 {
-    if (Tok == TOK_IDENT && strcmp (SVal, "unlimited") == 0) {
+    if (Tok == TOK_IDENT && SB_CompareStr (&SVal, "unlimited") == 0) {
 	NextTok ();
 	return -1;
     } else {
@@ -226,7 +258,7 @@ static long IntArg (long Min, long Max)
 
 
 
-static void ConDes (const char* Name, unsigned Type)
+static void ConDes (const StrBuf* Name, unsigned Type)
 /* Parse remaining line for constructor/destructor of the remaining type */
 {
     long Prio;
@@ -269,7 +301,7 @@ static void DoA16 (void)
 	Error ("Command is only valid in 65816 mode");
     } else {
        	/* Immidiate mode has two extension bytes */
-	ExtBytes [AMI_IMM_ACCU] = 2;
+	ExtBytes [AM65I_IMM_ACCU] = 2;
     }
 }
 
@@ -282,7 +314,7 @@ static void DoA8 (void)
 	Error ("Command is only valid in 65816 mode");
     } else {
 	/* Immidiate mode has one extension byte */
-	ExtBytes [AMI_IMM_ACCU] = 1;
+	ExtBytes [AM65I_IMM_ACCU] = 1;
     }
 }
 
@@ -349,8 +381,6 @@ static void DoAlign (void)
 static void DoASCIIZ (void)
 /* Define text with a zero terminator */
 {
-    unsigned Len;
-
     while (1) {
 	/* Must have a string constant */
 	if (Tok != TOK_STRCON) {
@@ -358,12 +388,9 @@ static void DoASCIIZ (void)
 	    return;
 	}
 
-	/* Get the length of the string constant */
-	Len = strlen (SVal);
-
 	/* Translate into target charset and emit */
-	TgtTranslateBuf (SVal, Len);
-       	EmitData ((unsigned char*) SVal, Len);
+       	TgtTranslateStrBuf (&SVal);
+       	EmitStrBuf (&SVal);
 	NextTok ();
 	if (Tok == TOK_COMMA) {
 	    NextTok ();
@@ -384,8 +411,8 @@ static void DoAssert (void)
         "ERROR"
     };
 
-    int Action;
-
+    int      Action;
+    unsigned Msg;
 
     /* First we have the expression that has to evaluated */
     ExprNode* Expr = Expression ();
@@ -414,15 +441,36 @@ static void DoAssert (void)
             Error ("Illegal assert action specifier");
     }
     NextTok ();
-    ConsumeComma ();
 
-    /* Read the message */
-    if (Tok != TOK_STRCON) {
-    	ErrorSkip ("String constant expected");
-    } else {
-        AddAssertion (Expr, Action, GetStringId (SVal));
+    /* We can have an optional message. If no message is present, use
+     * "Assertion failed".
+     */
+    if (Tok == TOK_COMMA) {
+
+        /* Skip the comma */
         NextTok ();
+
+        /* Read the message */
+        if (Tok != TOK_STRCON) {
+            ErrorSkip ("String constant expected");
+            return;
+        }
+
+        /* Translate the message into a string id. We can then skip the input
+         * string.
+         */
+        Msg = GetStrBufId (&SVal);
+        NextTok ();
+
+    } else {
+
+        /* Use "Assertion failed" */
+        Msg = GetStringId ("Assertion failed");
+
     }
+
+    /* Remember the assertion */
+    AddAssertion (Expr, Action, Msg);
 }
 
 
@@ -434,6 +482,20 @@ static void DoAutoImport (void)
 }
 
 
+static void DoBankBytes (void)
+/* Define bytes, extracting the bank byte from each expression in the list */
+{
+    while (1) {
+        EmitByte (FuncBankByte ());
+        if (Tok != TOK_COMMA) {
+            break;
+        } else {
+            NextTok ();
+        }
+    }
+}
+
+
 
 static void DoBss (void)
 /* Switch to the BSS segment */
@@ -449,14 +511,13 @@ static void DoByte (void)
     while (1) {
 	if (Tok == TOK_STRCON) {
 	    /* A string, translate into target charset and emit */
-	    unsigned Len = strlen (SVal);
-	    TgtTranslateBuf (SVal, Len);
-	    EmitData ((unsigned char*) SVal, Len);
+       	    TgtTranslateStrBuf (&SVal);
+	    EmitStrBuf (&SVal);
 	    NextTok ();
 	} else {
 	    EmitByte (Expression ());
 	}
-	if (Tok != TOK_COMMA) {
+    	if (Tok != TOK_COMMA) {
 	    break;
 	} else {
 	    NextTok ();
@@ -488,7 +549,7 @@ static void DoCharMap (void)
 
     /* Read the index as numerical value */
     Index = ConstExpression ();
-    if (Index < 1 || Index > 255) {
+    if (Index < 0 || Index > 255) {
       	/* Value out of range */
        	ErrorSkip ("Range error");
       	return;
@@ -499,10 +560,10 @@ static void DoCharMap (void)
 
     /* Read the character code */
     Code = ConstExpression ();
-    if (Code < 1 || Code > 255) {
-	/* Value out of range */
+    if (Code < 0 || Code > 255) {
+    	/* Value out of range */
        	ErrorSkip ("Range error");
-	return;
+    	return;
     }
 
     /* Set the character translation */
@@ -525,8 +586,9 @@ static void DoConDes (void)
     static const char* Keys[] = {
        	"CONSTRUCTOR",
 	"DESTRUCTOR",
+        "INTERRUPTOR",
     };
-    char Name [sizeof (SVal)];
+    StrBuf Name = STATIC_STRBUF_INITIALIZER;
     long Type;
 
     /* Symbol name follows */
@@ -534,7 +596,7 @@ static void DoConDes (void)
     	ErrorSkip ("Identifier expected");
     	return;
     }
-    strcpy (Name, SVal);
+    SB_Copy (&Name, &SVal);
     NextTok ();
 
     /* Type follows. May be encoded as identifier or numerical */
@@ -547,9 +609,8 @@ static void DoConDes (void)
 
 	/* Check if we got a valid keyword */
 	if (Type < 0) {
-	    Error ("Syntax error");
-	    SkipUntilSep ();
-	    return;
+       	    ErrorSkip ("Syntax error");
+	    goto ExitPoint;
 	}
 
     } else {
@@ -558,14 +619,18 @@ static void DoConDes (void)
        	Type = ConstExpression ();
     	if (Type < CD_TYPE_MIN || Type > CD_TYPE_MAX) {
     	    /* Value out of range */
-    	    Error ("Range error");
-    	    return;
+       	    ErrorSkip ("Range error");
+    	    goto ExitPoint;
     	}
 
     }
 
     /* Parse the remainder of the line and export the symbol */
-    ConDes (Name, (unsigned) Type);
+    ConDes (&Name, (unsigned) Type);
+
+ExitPoint:
+    /* Free string memory */
+    SB_Done (&Name);
 }
 
 
@@ -573,18 +638,21 @@ static void DoConDes (void)
 static void DoConstructor (void)
 /* Export a symbol as constructor */
 {
-    char Name [sizeof (SVal)];
+    StrBuf Name = STATIC_STRBUF_INITIALIZER;
 
     /* Symbol name follows */
     if (Tok != TOK_IDENT) {
     	ErrorSkip ("Identifier expected");
     	return;
     }
-    strcpy (Name, SVal);
+    SB_Copy (&Name, &SVal);
     NextTok ();
 
     /* Parse the remainder of the line and export the symbol */
-    ConDes (Name, CD_TYPE_CON);
+    ConDes (&Name, CD_TYPE_CON);
+
+    /* Free string memory */
+    SB_Done (&Name);
 }
 
 
@@ -665,18 +733,21 @@ static void DoDefine (void)
 static void DoDestructor (void)
 /* Export a symbol as destructor */
 {
-    char Name [sizeof (SVal)];
+    StrBuf Name = STATIC_STRBUF_INITIALIZER;
 
     /* Symbol name follows */
     if (Tok != TOK_IDENT) {
     	ErrorSkip ("Identifier expected");
     	return;
     }
-    strcpy (Name, SVal);
+    SB_Copy (&Name, &SVal);
     NextTok ();
 
     /* Parse the remainder of the line and export the symbol */
-    ConDes (Name, CD_TYPE_DES);
+    ConDes (&Name, CD_TYPE_DES);
+
+    /* Free string memory */
+    SB_Done (&Name);
 }
 
 
@@ -737,7 +808,7 @@ static void DoError (void)
     if (Tok != TOK_STRCON) {
  	ErrorSkip ("String constant expected");
     } else {
-       	Error ("User error: %s", SVal);
+       	Error ("User error: %m%p", &SVal);
 	SkipUntilSep ();
     }
 }
@@ -760,7 +831,7 @@ static void DoExitMacro (void)
 static void DoExport (void)
 /* Export a symbol */
 {
-    ExportImport (SymExport, ADDR_SIZE_DEFAULT, SF_NONE);
+    ExportImport (ExportWithAssign, ADDR_SIZE_DEFAULT, SF_NONE);
 }
 
 
@@ -768,7 +839,7 @@ static void DoExport (void)
 static void DoExportZP (void)
 /* Export a zeropage symbol */
 {
-    ExportImport (SymExport, ADDR_SIZE_ZP, SF_NONE);
+    ExportImport (ExportWithAssign, ADDR_SIZE_ZP, SF_NONE);
 }
 
 
@@ -804,9 +875,9 @@ static void DoFeature (void)
 	LocaseSVal ();
 
      	/* Set the feature and check for errors */
-	if (SetFeature (SVal) == FEAT_UNKNOWN) {
+	if (SetFeature (&SVal) == FEAT_UNKNOWN) {
      	    /* Not found */
-     	    ErrorSkip ("Invalid feature: `%s'", SVal);
+     	    ErrorSkip ("Invalid feature: `%m%p'", &SVal);
      	    return;
      	} else {
 	    /* Skip the keyword */
@@ -862,17 +933,17 @@ static void DoFileOpt (void)
 
 	    case 0:
 		/* Author */
-		OptAuthor (SVal);
+		OptAuthor (&SVal);
 		break;
 
 	    case 1:
 		/* Comment */
-		OptComment (SVal);
+		OptComment (&SVal);
 		break;
 
 	    case 2:
 		/* Compiler */
-		OptCompiler (SVal);
+		OptCompiler (&SVal);
 		break;
 
 	    default:
@@ -902,7 +973,7 @@ static void DoFileOpt (void)
 	}
 
 	/* Insert the option */
-	OptStr ((unsigned char) OptNum, SVal);
+	OptStr ((unsigned char) OptNum, &SVal);
 
 	/* Done */
 	NextTok ();
@@ -934,6 +1005,20 @@ static void DoGlobalZP (void)
 }
 
 
+static void DoHiBytes (void)
+/* Define bytes, extracting the hi byte from each expression in the list */
+{
+    while (1) {
+        EmitByte (FuncHiByte ());
+        if (Tok != TOK_COMMA) {
+            break;
+        } else {
+            NextTok ();
+        }
+    }
+}
+
+
 
 static void DoI16 (void)
 /* Switch the index registers to 16 bit mode (assembler only) */
@@ -942,7 +1027,7 @@ static void DoI16 (void)
      	Error ("Command is only valid in 65816 mode");
     } else {
        	/* Immidiate mode has two extension bytes */
-     	ExtBytes [AMI_IMM_INDEX] = 2;
+     	ExtBytes [AM65I_IMM_INDEX] = 2;
     }
 }
 
@@ -955,7 +1040,7 @@ static void DoI8 (void)
 	Error ("Command is only valid in 65816 mode");
     } else {
 	/* Immidiate mode has one extension byte */
-	ExtBytes [AMI_IMM_INDEX] = 1;
+	ExtBytes [AM65I_IMM_INDEX] = 1;
     }
 }
 
@@ -980,7 +1065,7 @@ static void DoImportZP (void)
 static void DoIncBin (void)
 /* Include a binary file */
 {
-    char Name [sizeof (SVal)];
+    StrBuf Name = STATIC_STRBUF_INITIALIZER;
     long Start = 0L;
     long Count = -1L;
     long Size;
@@ -991,7 +1076,8 @@ static void DoIncBin (void)
     	ErrorSkip ("String constant expected");
     	return;
     }
-    strcpy (Name, SVal);
+    SB_Copy (&Name, &SVal);
+    SB_Terminate (&Name);
     NextTok ();
 
     /* A starting offset may follow */
@@ -1008,14 +1094,14 @@ static void DoIncBin (void)
     }
 
     /* Try to open the file */
-    F = fopen (Name, "rb");
+    F = fopen (SB_GetConstBuf (&Name), "rb");
     if (F == 0) {
 
        	/* Search for the file in the include directories. */
-     	char* PathName = FindInclude (Name);
-       	if (PathName == 0 || (F = fopen (PathName, "r")) == 0) {
+     	char* PathName = FindInclude (SB_GetConstBuf (&Name));
+       	if (PathName == 0 || (F = fopen (PathName, "rb")) == 0) {
      	    /* Not found or cannot open, print an error and bail out */
-       	    ErrorSkip ("Cannot open include file `%s': %s", Name, strerror (errno));
+       	    ErrorSkip ("Cannot open include file `%m%p': %s", &Name, strerror (errno));
      	}
 
      	/* Free the allocated memory */
@@ -1023,7 +1109,7 @@ static void DoIncBin (void)
 
         /* If we had an error before, bail out now */
         if (F == 0) {
-            return;
+            goto ExitPoint;
         }
     }
 
@@ -1062,8 +1148,8 @@ static void DoIncBin (void)
 	size_t BytesRead = fread (Buf, 1, BytesToRead, F);
 	if (BytesToRead != BytesRead) {
 	    /* Some sort of error */
-	    ErrorSkip ("Cannot read from include file `%s': %s",
-                       Name, strerror (errno));
+	    ErrorSkip ("Cannot read from include file `%m%p': %s",
+                       &Name, strerror (errno));
 	    break;
 	}
 
@@ -1077,6 +1163,10 @@ static void DoIncBin (void)
 Done:
     /* Close the file, ignore errors since it's r/o */
     (void) fclose (F);
+
+ExitPoint:
+    /* Free string memory */
+    SB_Done (&Name);
 }
 
 
@@ -1084,20 +1174,42 @@ Done:
 static void DoInclude (void)
 /* Include another file */
 {
-    char Name [MAX_STR_LEN+1];
-
     /* Name must follow */
     if (Tok != TOK_STRCON) {
 	ErrorSkip ("String constant expected");
     } else {
-    	strcpy (Name, SVal);
-    	NextTok ();
-    	NewInputFile (Name);
+        SB_Terminate (&SVal);
+    	if (NewInputFile (SB_GetConstBuf (&SVal)) == 0) {
+            /* Error opening the file, skip remainder of line */
+            SkipUntilSep ();
+        }
     }
 }
 
 
 
+static void DoInterruptor (void)
+/* Export a symbol as interruptor */
+{
+    StrBuf Name = STATIC_STRBUF_INITIALIZER;
+
+    /* Symbol name follows */
+    if (Tok != TOK_IDENT) {
+    	ErrorSkip ("Identifier expected");
+    	return;
+    }
+    SB_Copy (&Name, &SVal);
+    NextTok ();
+
+    /* Parse the remainder of the line and export the symbol */
+    ConDes (&Name, CD_TYPE_INT);
+
+    /* Free string memory */
+    SB_Done (&Name);
+}
+
+
+
 static void DoInvalid (void)
 /* Handle a token that is invalid here, since it should have been handled on
  * a much lower level of the expression hierarchy. Getting this sort of token
@@ -1107,7 +1219,7 @@ static void DoInvalid (void)
  * an error in the assembler itself, while DoInvalid is.
  */
 {
-    Internal ("Unexpected token: %s", Keyword);
+    Internal ("Unexpected token: %m%p", &Keyword);
 }
 
 
@@ -1137,6 +1249,20 @@ static void DoList (void)
 
 
 
+static void DoLoBytes (void)
+/* Define bytes, extracting the lo byte from each expression in the list */
+{
+    while (1) {
+        EmitByte (FuncLoByte ());
+        if (Tok != TOK_COMMA) {
+            break;
+        } else {
+            NextTok ();
+        }
+    }
+}
+
+
 static void DoListBytes (void)
 /* Set maximum number of bytes to list for one line */
 {
@@ -1165,14 +1291,6 @@ static void DoLocalChar (void)
 static void DoMacPack (void)
 /* Insert a macro package */
 {
-    /* Macro package names */
-    static const char* Keys [] = {
-    	"GENERIC",
-       	"LONGBRANCH",
-        "CBM",
-        "CPU"
-    };
-
     int Package;
 
     /* We expect an identifier */
@@ -1181,19 +1299,21 @@ static void DoMacPack (void)
     	return;
     }
 
-    /* Map the keyword to a number */
-    Package = GetSubKey (Keys, sizeof (Keys) / sizeof (Keys [0]));
+    /* Search for the macro package name */
+    LocaseSVal ();
+    Package = MacPackFind (&SVal);
     if (Package < 0) {
     	/* Not found */
     	ErrorSkip ("Invalid macro package");
     	return;
     }
 
-    /* Skip the package name */
-    NextTok ();
-
-    /* Insert the package */
-    InsertMacPack (Package);
+    /* Insert the package. If this fails, skip the remainder of the line to
+     * avoid additional error messages.
+     */
+    if (MacPackInsert (Package) == 0) {
+        SkipUntilSep ();
+    }
 }
 
 
@@ -1222,7 +1342,7 @@ static void DoOrg (void)
 	Error ("Range error");
     	return;
     }
-    SetAbsPC (PC);
+    EnterAbsoluteMode (PC);
 }
 
 
@@ -1236,7 +1356,7 @@ static void DoOut (void)
 	/* Output the string and be sure to flush the output to keep it in
 	 * sync with any error messages if the output is redirected to a file.
 	 */
-	printf ("%s\n", SVal);
+	printf ("%.*s\n", (int) SB_GetLen (&SVal), SB_GetConstBuf (&SVal));
 	fflush (stdout);
 	NextTok ();
     }
@@ -1302,7 +1422,7 @@ static void DoPopSeg (void)
 static void DoProc (void)
 /* Start a new lexical scope */
 {
-    char Name[sizeof(SVal)];
+    StrBuf Name = STATIC_STRBUF_INITIALIZER;
     unsigned char AddrSize;
 
     if (Tok == TOK_IDENT) {
@@ -1310,10 +1430,10 @@ static void DoProc (void)
         SymEntry* Sym;
 
 	/* The new scope has a name. Remember it. */
-        strcpy (Name, SVal);
+        SB_Copy (&Name, &SVal);
 
         /* Search for the symbol, generate a new one if needed */
-       	Sym = SymFind (CurrentScope, Name, SYM_ALLOC_NEW);
+       	Sym = SymFind (CurrentScope, &Name, SYM_ALLOC_NEW);
 
         /* Skip the scope name */
         NextTok ();
@@ -1328,13 +1448,16 @@ static void DoProc (void)
 
         /* A .PROC statement without a name */
         Warning (1, "Unnamed .PROCs are deprecated, please use .SCOPE");
-        AnonName (Name, sizeof (Name), "PROC");
+        AnonName (&Name, "PROC");
         AddrSize = ADDR_SIZE_DEFAULT;
 
     }
 
     /* Enter a new scope */
-    SymEnterLevel (Name, ST_PROC, AddrSize);
+    SymEnterLevel (&Name, ST_PROC, AddrSize);
+
+    /* Free memory for Name */
+    SB_Done (&Name);
 }
 
 
@@ -1365,7 +1488,7 @@ static void DoPushSeg (void)
 static void DoReloc (void)
 /* Enter relocatable mode */
 {
-    RelocMode = 1;
+    EnterRelocMode ();
 }
 
 
@@ -1422,20 +1545,20 @@ static void DoROData (void)
 static void DoScope (void)
 /* Start a local scope */
 {
-    char Name[sizeof (SVal)];
+    StrBuf Name = STATIC_STRBUF_INITIALIZER;
     unsigned char AddrSize;
 
 
     if (Tok == TOK_IDENT) {
 
     	/* The new scope has a name. Remember and skip it. */
-        strcpy (Name, SVal);
+        SB_Copy (&Name, &SVal);
         NextTok ();
 
     } else {
 
         /* An unnamed scope */
-        AnonName (Name, sizeof (Name), "SCOPE");
+        AnonName (&Name, "SCOPE");
 
     }
 
@@ -1443,8 +1566,10 @@ static void DoScope (void)
     AddrSize = OptionalAddrSize ();
 
     /* Enter the new scope */
-    SymEnterLevel (Name, ST_SCOPE, AddrSize);
+    SymEnterLevel (&Name, ST_SCOPE, AddrSize);
 
+    /* Free memory for Name */
+    SB_Done (&Name);
 }
 
 
@@ -1452,24 +1577,30 @@ static void DoScope (void)
 static void DoSegment (void)
 /* Switch to another segment */
 {
-    char Name [sizeof (SVal)];
+    StrBuf Name = STATIC_STRBUF_INITIALIZER;
     SegDef Def;
-    Def.Name = Name;
 
     if (Tok != TOK_STRCON) {
 	ErrorSkip ("String constant expected");
     } else {
 
 	/* Save the name of the segment and skip it */
-	strcpy (Name, SVal);
+	SB_Copy (&Name, &SVal);
 	NextTok ();
 
+        /* Use the name for the segment definition */
+        SB_Terminate (&Name);
+        Def.Name = SB_GetBuf (&Name);
+
     	/* Check for an optional address size modifier */
         Def.AddrSize = OptionalAddrSize ();
 
 	/* Set the segment */
      	UseSeg (&Def);
     }
+
+    /* Free memory for Name */
+    SB_Done (&Name);
 }
 
 
@@ -1481,12 +1612,19 @@ static void DoSetCPU (void)
     if (Tok != TOK_STRCON) {
 	ErrorSkip ("String constant expected");
     } else {
-        /* Try to find the CPU, then skip the identifier */
-        cpu_t CPU = FindCPU (SVal);
-        NextTok ();
+        cpu_t CPU;
+
+        /* Try to find the CPU */
+        SB_Terminate (&SVal);
+        CPU = FindCPU (SB_GetConstBuf (&SVal));
 
         /* Switch to the new CPU */
         SetCPU (CPU);
+
+        /* Skip the identifier. If the CPU switch was successful, the scanner
+         * will treat the input now correctly for the new CPU.
+         */
+        NextTok ();
     }
 }
 
@@ -1558,7 +1696,7 @@ static void DoTag (void)
 static void DoUnexpected (void)
 /* Got an unexpected keyword */
 {
-    Error ("Unexpected `%s'", Keyword);
+    Error ("Unexpected `%m%p'", &Keyword);
     SkipUntilSep ();
 }
 
@@ -1570,7 +1708,7 @@ static void DoWarning (void)
     if (Tok != TOK_STRCON) {
  	ErrorSkip ("String constant expected");
     } else {
-       	Warning (0, "User warning: %s", SVal);
+       	Warning (0, "User warning: %m%p", &SVal);
 	SkipUntilSep ();
     }
 }
@@ -1629,6 +1767,7 @@ static CtrlDesc CtrlCmdTab [] = {
     { ccNone,           DoAssert        },
     { ccNone,       	DoAutoImport	},
     { ccNone,        	DoUnexpected	},	/* .BANKBYTE */
+    { ccNone,           DoBankBytes     },
     { ccNone,        	DoUnexpected	},	/* .BLANK */
     { ccNone,       	DoBss		},
     { ccNone,       	DoByte		},
@@ -1672,9 +1811,11 @@ static CtrlDesc CtrlCmdTab [] = {
     { ccNone,		DoGlobal	},
     { ccNone,		DoGlobalZP	},
     { ccNone,        	DoUnexpected	},	/* .HIBYTE */
+    { ccNone,           DoHiBytes       },
     { ccNone,        	DoUnexpected	},	/* .HIWORD */
     { ccNone,		DoI16	  	},
     { ccNone,		DoI8	  	},
+    { ccNone,           DoUnexpected    },      /* .IDENT */
     { ccKeepToken,	DoConditionals	},	/* .IF */
     { ccKeepToken,	DoConditionals	},	/* .IFBLANK */
     { ccKeepToken,	DoConditionals	},	/* .IFCONST */
@@ -1692,11 +1833,13 @@ static CtrlDesc CtrlCmdTab [] = {
     { ccNone,		DoImportZP	},
     { ccNone,		DoIncBin	},
     { ccNone,      	DoInclude	},
+    { ccNone,           DoInterruptor   },
     { ccNone,		DoInvalid	},	/* .LEFT */
     { ccNone,		DoLineCont	},
     { ccNone,		DoList		},
     { ccNone,  	       	DoListBytes	},
     { ccNone,        	DoUnexpected	},	/* .LOBYTE */
+    { ccNone,           DoLoBytes       },
     { ccNone,		DoUnexpected	},	/* .LOCAL */
     { ccNone,		DoLocalChar	},
     { ccNone,        	DoUnexpected	},	/* .LOWORD */
@@ -1724,9 +1867,11 @@ static CtrlDesc CtrlCmdTab [] = {
     { ccNone,		DoROData	},
     { ccNone,           DoScope         },
     { ccNone,       	DoSegment	},
+    { ccNone,           DoUnexpected    },      /* .SET */
     { ccNone,          	DoSetCPU        },
     { ccNone,           DoUnexpected    },      /* .SIZEOF */
     { ccNone,        	DoSmart		},
+    { ccNone,           DoUnexpected    },      /* .SPRINTF */
     { ccNone,		DoUnexpected	},   	/* .STRAT */
     { ccNone,          	DoUnexpected	},   	/* .STRING */
     { ccNone,		DoUnexpected	},   	/* .STRLEN */
@@ -1762,7 +1907,7 @@ void HandlePseudo (void)
     /* Safety check */
     if (PSEUDO_COUNT != (TOK_LASTPSEUDO - TOK_FIRSTPSEUDO + 1)) {
 	Internal ("Pseudo mismatch: PSEUDO_COUNT = %u, actual count = %u\n",
-		  PSEUDO_COUNT, TOK_LASTPSEUDO - TOK_FIRSTPSEUDO + 1);
+       	       	  (unsigned) PSEUDO_COUNT, TOK_LASTPSEUDO - TOK_FIRSTPSEUDO + 1);
     }
     CHECK (Index < PSEUDO_COUNT);
 
@@ -1771,7 +1916,7 @@ void HandlePseudo (void)
 
     /* Remember the instruction, then skip it if needed */
     if ((D->Flags & ccKeepToken) == 0) {
-    	strcpy (Keyword, SVal);
+    	SB_Copy (&Keyword, &SVal);
      	NextTok ();
     }
 
