@@ -339,8 +339,12 @@ static unsigned ParseStaticDecl (Declaration* Decl, unsigned* SC)
     /* Static data */
     if (CurTok.Tok == TOK_ASSIGN) {
 
-        /* Initialization ahead, switch to data segment */
-        if (IsQualConst (Decl->Type)) {
+        /* Initialization ahead, switch to data segment. For arrays, we need 
+         * to check the elements of the array for constness, not the array 
+         * itself.
+         */
+        if (IsQualConst (Decl->Type) ||
+            (IsTypeArray (Decl->Type) && IsQualConst (GetElementType (Decl->Type)))) {
             g_userodata ();
         } else {
             g_usedata ();
