@@ -448,7 +448,7 @@ void Convert (const O65Data* D)
     }
 
     /* Open the output file */
-    F = fopen (OutputName, "wb");
+    F = fopen (OutputName, "w");
     if (F == 0) {
         Error ("Cannot open `%s': %s", OutputName, strerror (errno));
     }
