@@ -6,8 +6,8 @@
 /*                                                                           */
 /*                                                                           */
 /*                                                                           */
-/* (C) 2000-2003 Ullrich von Bassewitz                                       */
-/*               Rmerstrasse 52                                             */
+/* (C) 2000-2007 Ullrich von Bassewitz                                       */
+/*               Roemerstrasse 52                                            */
 /*               D-70794 Filderstadt                                         */
 /* EMail:        uz@cc65.org                                                 */
 /*                                                                           */
@@ -35,6 +35,7 @@
 
 #include <stdio.h>
 #include <string.h>
+#include <limits.h>
 #if defined(_MSC_VER)
 /* Microsoft compiler */
 #  include <io.h>
@@ -48,12 +49,16 @@
 #include "xmalloc.h"
 
 /* da65 */
+#include "asminc.h"
 #include "attrtab.h"
+#include "comments.h"
 #include "error.h"
 #include "global.h"
 #include "infofile.h"
+#include "labels.h"
 #include "opctable.h"
 #include "scanner.h"
+#include "segment.h"
 
 
 
@@ -75,19 +80,121 @@ static void AddAttr (const char* Name, unsigned* Set, unsigned Attr)
 
 
 
+static void AsmIncSection (void)
+/* Parse a asminc section */
+{
+    static const IdentTok LabelDefs[] = {
+        {   "COMMENTSTART",     INFOTOK_COMMENTSTART    },
+       	{   "FILE",    	        INFOTOK_FILE            },                     
+        {   "IGNOREUNKNOWN",    INFOTOK_IGNOREUNKNOWN   },
+    };
+
+    /* Locals - initialize to avoid gcc warnings */
+    char* Name = 0;
+    int CommentStart = EOF;
+    int IgnoreUnknown = -1;
+
+    /* Skip the token */
+    InfoNextTok ();
+
+    /* Expect the opening curly brace */
+    InfoConsumeLCurly ();
+
+    /* Look for section tokens */
+    while (InfoTok != INFOTOK_RCURLY) {
+
+	/* Convert to special token */
+       	InfoSpecialToken (LabelDefs, ENTRY_COUNT (LabelDefs), "Asminc directive");
+
+	/* Look at the token */
+	switch (InfoTok) {
+
+            case INFOTOK_COMMENTSTART:
+                InfoNextTok ();
+                if (CommentStart != EOF) {
+                    InfoError ("Commentstart already given");
+                }
+                InfoAssureChar ();
+                CommentStart = (char) InfoIVal;
+                InfoNextTok ();
+                break;
+
+	    case INFOTOK_FILE:
+	        InfoNextTok ();
+	       	if (Name) {
+	       	    InfoError ("File name already given");
+	       	}
+	       	InfoAssureStr ();
+		if (InfoSVal[0] == '\0') {
+		    InfoError ("File name may not be empty");
+		}
+	       	Name = xstrdup (InfoSVal);
+	       	InfoNextTok ();
+	       	break;
+
+            case INFOTOK_IGNOREUNKNOWN:
+                InfoNextTok ();
+                if (IgnoreUnknown != -1) {
+                    InfoError ("Ignoreunknown already specified");
+                }
+                InfoBoolToken ();
+                IgnoreUnknown = (InfoTok != INFOTOK_FALSE);
+                InfoNextTok ();
+                break;
+
+            default:
+                Internal ("Unexpected token: %u", InfoTok);
+	}
+
+	/* Directive is followed by a semicolon */
+	InfoConsumeSemi ();
+    }
+
+    /* Check for the necessary data and assume defaults */
+    if (Name == 0) {
+	InfoError ("File name is missing");
+    }
+    if (CommentStart == EOF) {
+	CommentStart = ';';
+    }
+    if (IgnoreUnknown == -1) {
+        IgnoreUnknown = 0;
+    }
+
+    /* Open the file and read the symbol definitions */
+    AsmInc (Name, CommentStart, IgnoreUnknown);
+
+    /* Delete the dynamically allocated memory for Name */
+    xfree (Name);
+
+    /* Consume the closing brace */
+    InfoConsumeRCurly ();
+}
+
+
+
 static void GlobalSection (void)
 /* Parse a global section */
 {
     static const IdentTok GlobalDefs[] = {
+        {   "ARGUMENTCOL",      INFOTOK_ARGUMENT_COLUMN },
+        {   "ARGUMENTCOLUMN",   INFOTOK_ARGUMENT_COLUMN },
+        {   "COMMENTCOL",       INFOTOK_COMMENT_COLUMN  },
+        {   "COMMENTCOLUMN",    INFOTOK_COMMENT_COLUMN  },
         {   "COMMENTS",         INFOTOK_COMMENTS        },
        	{   "CPU",     	        INFOTOK_CPU     	},
         {   "HEXOFFS",          INFOTOK_HEXOFFS         },
-       	{   "INPUTNAME",  	INFOTOK_INPUTNAME	},
+       	{   "INPUTNAME",    	INFOTOK_INPUTNAME	},
         {   "INPUTOFFS",        INFOTOK_INPUTOFFS       },
         {   "INPUTSIZE",        INFOTOK_INPUTSIZE       },
+        {   "LABELBREAK",       INFOTOK_LABELBREAK      },
+        {   "MNEMONICCOL",      INFOTOK_MNEMONIC_COLUMN },
+        {   "MNEMONICCOLUMN",   INFOTOK_MNEMONIC_COLUMN },
 	{   "OUTPUTNAME",      	INFOTOK_OUTPUTNAME	},
 	{   "PAGELENGTH",      	INFOTOK_PAGELENGTH	},
-	{   "STARTADDR",	INFOTOK_STARTADDR	},
+	{   "STARTADDR",     	INFOTOK_STARTADDR	},
+        {   "TEXTCOL",          INFOTOK_TEXT_COLUMN     },
+        {   "TEXTCOLUMN",       INFOTOK_TEXT_COLUMN     },
     };
 
     /* Skip the token */
@@ -105,6 +212,22 @@ static void GlobalSection (void)
 	/* Look at the token */
 	switch (InfoTok) {
 
+            case INFOTOK_ARGUMENT_COLUMN:
+		InfoNextTok ();
+		InfoAssureInt ();
+                InfoRangeCheck (MIN_ACOL, MAX_ACOL);
+	     	ACol = InfoIVal;
+		InfoNextTok ();
+		break;
+
+            case INFOTOK_COMMENT_COLUMN:
+		InfoNextTok ();
+		InfoAssureInt ();
+                InfoRangeCheck (MIN_CCOL, MAX_CCOL);
+	     	CCol = InfoIVal;
+		InfoNextTok ();
+		break;
+
             case INFOTOK_COMMENTS:
 		InfoNextTok ();
 		InfoAssureInt ();
@@ -159,6 +282,22 @@ static void GlobalSection (void)
 		InfoNextTok ();
 		break;
 
+            case INFOTOK_LABELBREAK:
+		InfoNextTok ();
+		InfoAssureInt ();
+                InfoRangeCheck (0, UCHAR_MAX);
+		LBreak = (unsigned char) InfoIVal;
+		InfoNextTok ();
+		break;
+
+            case INFOTOK_MNEMONIC_COLUMN:
+		InfoNextTok ();
+		InfoAssureInt ();
+                InfoRangeCheck (MIN_MCOL, MAX_MCOL);
+	     	MCol = InfoIVal;
+		InfoNextTok ();
+		break;
+
 	    case INFOTOK_OUTPUTNAME:
 		InfoNextTok ();
 		InfoAssureStr ();
@@ -187,6 +326,17 @@ static void GlobalSection (void)
 		InfoNextTok ();
 		break;
 
+            case INFOTOK_TEXT_COLUMN:
+		InfoNextTok ();
+		InfoAssureInt ();
+                InfoRangeCheck (MIN_TCOL, MAX_TCOL);
+	     	TCol = InfoIVal;
+		InfoNextTok ();
+		break;
+
+            default:
+                Internal ("Unexpected token: %u", InfoTok);
+
 	}
 
 	/* Directive is followed by a semicolon */
@@ -200,10 +350,139 @@ static void GlobalSection (void)
 
 
 
+static void LabelSection (void)
+/* Parse a label section */
+{
+    static const IdentTok LabelDefs[] = {
+       	{   "COMMENT",  INFOTOK_COMMENT },
+	{   "ADDR",	INFOTOK_ADDR	},
+       	{   "NAME",	INFOTOK_NAME	},
+       	{   "SIZE",    	INFOTOK_SIZE	},
+    };
+
+    /* Locals - initialize to avoid gcc warnings */
+    char* Name    = 0;
+    char* Comment = 0;
+    long Value    = -1;
+    long Size     = -1;
+
+    /* Skip the token */
+    InfoNextTok ();
+
+    /* Expect the opening curly brace */
+    InfoConsumeLCurly ();
+
+    /* Look for section tokens */
+    while (InfoTok != INFOTOK_RCURLY) {
+
+	/* Convert to special token */
+       	InfoSpecialToken (LabelDefs, ENTRY_COUNT (LabelDefs), "Label attribute");
+
+	/* Look at the token */
+	switch (InfoTok) {
+
+	    case INFOTOK_ADDR:
+	       	InfoNextTok ();
+	       	if (Value >= 0) {
+	       	    InfoError ("Value already given");
+	       	}
+     	       	InfoAssureInt ();
+		InfoRangeCheck (0, 0xFFFF);
+	       	Value = InfoIVal;
+	       	InfoNextTok ();
+	       	break;
+
+	    case INFOTOK_COMMENT:
+	        InfoNextTok ();
+	       	if (Comment) {
+	       	    InfoError ("Comment already given");
+	       	}
+	       	InfoAssureStr ();
+		if (InfoSVal[0] == '\0') {
+		    InfoError ("Comment may not be empty");
+		}
+	       	Comment = xstrdup (InfoSVal);
+	       	InfoNextTok ();
+	       	break;
+
+	    case INFOTOK_NAME:
+	        InfoNextTok ();
+	       	if (Name) {
+	       	    InfoError ("Name already given");
+	       	}
+     	       	InfoAssureStr ();
+	       	Name = xstrdup (InfoSVal);
+	       	InfoNextTok ();
+	       	break;
+
+	    case INFOTOK_SIZE:
+	       	InfoNextTok ();
+	       	if (Size >= 0) {
+	       	    InfoError ("Size already given");
+	       	}
+	       	InfoAssureInt ();
+	    	InfoRangeCheck (1, 0x10000);
+	       	Size = InfoIVal;
+	       	InfoNextTok ();
+     	       	break;
+
+            default:
+                Internal ("Unexpected token: %u", InfoTok);
+     	}
+
+     	/* Directive is followed by a semicolon */
+     	InfoConsumeSemi ();
+    }
+
+    /* Did we get the necessary data */
+    if (Name == 0) {
+     	InfoError ("Label name is missing");
+    }
+    if (Name[0] == '\0' && Size > 1) {
+        InfoError ("Unnamed labels must not have a size > 1");
+    }
+    if (Value < 0) {
+	InfoError ("Label value is missing");
+    }
+    if (Size < 0) {
+	/* Use default */
+	Size = 1;
+    }
+    if (Value + Size > 0x10000) {
+    	InfoError ("Invalid size (address out of range)");
+    }
+    if (HaveLabel ((unsigned) Value)) {
+    	InfoError ("Label for address $%04lX already defined", Value);
+    }
+
+    /* Define the label(s) */
+    if (Name[0] == '\0') {
+        /* Size has already beed checked */
+        AddUnnamedLabel (Value);
+    } else {
+        AddExtLabelRange ((unsigned) Value, Name, Size);
+    }
+
+    /* Define the comment */
+    if (Comment) {
+        SetComment (Value, Comment);
+    }
+
+    /* Delete the dynamically allocated memory for Name and Comment */
+    xfree (Name);
+    xfree (Comment);
+
+    /* Consume the closing brace */
+    InfoConsumeRCurly ();
+}
+
+
+
 static void RangeSection (void)
 /* Parse a range section */
 {
     static const IdentTok RangeDefs[] = {
+        {   "COMMENT",          INFOTOK_COMMENT },
 	{   "END",	    	INFOTOK_END 	},
         {   "NAME",             INFOTOK_NAME    },
        	{   "START",   	    	INFOTOK_START	},
@@ -211,11 +490,11 @@ static void RangeSection (void)
     };
 
     static const IdentTok TypeDefs[] = {
-      	{   "ADDRTABLE",	INFOTOK_ADDRTAB	 },
+      	{   "ADDRTABLE",    	INFOTOK_ADDRTAB	 },
       	{   "BYTETABLE",    	INFOTOK_BYTETAB	 },
       	{   "CODE",	    	INFOTOK_CODE	 },
         {   "DBYTETABLE",       INFOTOK_DBYTETAB },
-      	{   "DWORDTABLE",	INFOTOK_DWORDTAB },
+      	{   "DWORDTABLE",   	INFOTOK_DWORDTAB },
       	{   "RTSTABLE",	    	INFOTOK_RTSTAB	 },
         {   "SKIP",             INFOTOK_SKIP     },
       	{   "TEXTTABLE",        INFOTOK_TEXTTAB  },
@@ -230,6 +509,7 @@ static void RangeSection (void)
       	tEnd	= 0x02,
       	tType	= 0x04,
         tName   = 0x08,
+        tComment= 0x10,
        	tNeeded = (tStart | tEnd | tType)
     };
     unsigned Attributes = tNone;
@@ -239,6 +519,9 @@ static void RangeSection (void)
     unsigned End	= 0;
     unsigned char Type	= 0;
     char* Name          = 0;
+    char* Comment       = 0;
+    unsigned MemberSize = 0;
+
 
     /* Skip the token */
     InfoNextTok ();
@@ -250,11 +533,23 @@ static void RangeSection (void)
     while (InfoTok != INFOTOK_RCURLY) {
 
 	/* Convert to special token */
-       	InfoSpecialToken (RangeDefs, ENTRY_COUNT (RangeDefs), "Range directive");
+       	InfoSpecialToken (RangeDefs, ENTRY_COUNT (RangeDefs), "Range attribute");
 
 	/* Look at the token */
 	switch (InfoTok) {
 
+	    case INFOTOK_COMMENT:
+                AddAttr ("COMMENT", &Attributes, tComment);
+	        InfoNextTok ();
+	       	InfoAssureStr ();
+	     	if (InfoSVal[0] == '\0') {
+	     	    InfoError ("Comment may not be empty");
+	     	}
+	       	Comment = xstrdup (InfoSVal);
+                Attributes |= tComment;
+	       	InfoNextTok ();
+      	       	break;
+
 	    case INFOTOK_END:
                 AddAttr ("END", &Attributes, tEnd);
 	     	InfoNextTok ();
@@ -290,18 +585,21 @@ static void RangeSection (void)
 	     	InfoNextTok ();
 		InfoSpecialToken (TypeDefs, ENTRY_COUNT (TypeDefs), "TYPE");
 		switch (InfoTok) {
-		    case INFOTOK_ADDRTAB:	Type = atAddrTab;	break;
-		    case INFOTOK_BYTETAB:	Type = atByteTab;	break;
-		    case INFOTOK_CODE:		Type = atCode;		break;
-                    case INFOTOK_DBYTETAB:      Type = atDByteTab;      break;
-		    case INFOTOK_DWORDTAB:	Type = atDWordTab;	break;
-       		    case INFOTOK_RTSTAB:       	Type = atRtsTab;	break;
-                    case INFOTOK_SKIP:          Type = atSkip;          break;
-		    case INFOTOK_TEXTTAB:       Type = atTextTab;       break;
-		    case INFOTOK_WORDTAB:	Type = atWordTab;	break;
+		    case INFOTOK_ADDRTAB:  Type = atAddrTab;  MemberSize = 2; break;
+		    case INFOTOK_BYTETAB:  Type = atByteTab;  MemberSize = 1; break;
+		    case INFOTOK_CODE:	   Type = atCode;     MemberSize = 1; break;
+                    case INFOTOK_DBYTETAB: Type = atDByteTab; MemberSize = 2; break;
+		    case INFOTOK_DWORDTAB: Type = atDWordTab; MemberSize = 4; break;
+       		    case INFOTOK_RTSTAB:   Type = atRtsTab;   MemberSize = 2; break;
+                    case INFOTOK_SKIP:     Type = atSkip;     MemberSize = 1; break;
+		    case INFOTOK_TEXTTAB:  Type = atTextTab;  MemberSize = 1; break;
+		    case INFOTOK_WORDTAB:  Type = atWordTab;  MemberSize = 2; break;
 		}
 		InfoNextTok ();
 		break;
+
+            default:
+                Internal ("Unexpected token: %u", InfoTok);
 	}
 
 	/* Directive is followed by a semicolon */
@@ -319,15 +617,28 @@ static void RangeSection (void)
 	InfoError ("Start value must not be greater than end value");
     }
 
+    /* Check the granularity */
+    if (((End - Start + 1) % MemberSize) != 0) {
+	InfoError ("Type of range needs a granularity of %u", MemberSize);
+    }
+
     /* Set the range */
     MarkRange (Start, End, Type);
 
     /* Do we have a label? */
     if (Attributes & tName) {
+
         /* Define a label for the table */
-        AddLabel (Start, atExtLabel, Name);
-        /* Delete the name */
+        AddExtLabel (Start, Name);
+
+        /* Set the comment if we have one */
+        if (Comment) {
+            SetComment (Start, Comment);
+        }
+
+        /* Delete name and comment */
         xfree (Name);
+        xfree (Comment);
     }
 
     /* Consume the closing brace */
@@ -336,19 +647,19 @@ static void RangeSection (void)
 
 
 
-static void LabelSection (void)
-/* Parse a label section */
+static void SegmentSection (void)
+/* Parse a segment section */
 {
     static const IdentTok LabelDefs[] = {
+       	{   "END",     	INFOTOK_END     },
        	{   "NAME",	INFOTOK_NAME	},
-	{   "ADDR",	INFOTOK_ADDR	},
-       	{   "SIZE",    	INFOTOK_SIZE	},
+	{   "START",    INFOTOK_START   },
     };
 
     /* Locals - initialize to avoid gcc warnings */
-    char* Name = 0;
-    long Value = -1;
-    long Size  = -1;
+    long End    = -1;
+    long Start  = -1;
+    char* Name  = 0;
 
     /* Skip the token */
     InfoNextTok ();
@@ -359,73 +670,76 @@ static void LabelSection (void)
     /* Look for section tokens */
     while (InfoTok != INFOTOK_RCURLY) {
 
-	/* Convert to special token */
-       	InfoSpecialToken (LabelDefs, ENTRY_COUNT (LabelDefs), "Label directive");
+    	/* Convert to special token */
+       	InfoSpecialToken (LabelDefs, ENTRY_COUNT (LabelDefs), "Segment attribute");
 
-	/* Look at the token */
-	switch (InfoTok) {
+    	/* Look at the token */
+    	switch (InfoTok) {
+
+    	    case INFOTOK_END:
+    	       	InfoNextTok ();
+    	       	if (End >= 0) {
+    	       	    InfoError ("Value already given");
+    	       	}
+     	       	InfoAssureInt ();
+    		InfoRangeCheck (0, 0xFFFF);
+    	       	End = InfoIVal;
+    	       	InfoNextTok ();
+	       	break;
 
 	    case INFOTOK_NAME:
 	        InfoNextTok ();
 	       	if (Name) {
 	       	    InfoError ("Name already given");
 	       	}
-	       	InfoAssureStr ();
-		if (InfoSVal[0] == '\0') {
-		    InfoError ("Name may not be empty");
-		}
+     	       	InfoAssureStr ();
 	       	Name = xstrdup (InfoSVal);
 	       	InfoNextTok ();
 	       	break;
 
-	    case INFOTOK_ADDR:
-	       	InfoNextTok ();
-	       	if (Value >= 0) {
-	       	    InfoError ("Value already given");
-	       	}
-	       	InfoAssureInt ();
-		InfoRangeCheck (0, 0xFFFF);
-	       	Value = InfoIVal;
-	       	InfoNextTok ();
+    	    case INFOTOK_START:
+    	       	InfoNextTok ();
+    	       	if (Start >= 0) {
+    	       	    InfoError ("Value already given");
+    	       	}
+     	       	InfoAssureInt ();
+    		InfoRangeCheck (0, 0xFFFF);
+    	       	Start = InfoIVal;
+    	       	InfoNextTok ();
 	       	break;
 
-	    case INFOTOK_SIZE:
-	       	InfoNextTok ();
-	       	if (Size >= 0) {
-	       	    InfoError ("Size already given");
-	       	}
-	       	InfoAssureInt ();
-		InfoRangeCheck (1, 0x10000);
-	       	Size = InfoIVal;
-	       	InfoNextTok ();
-	       	break;
+            default:
+                Internal ("Unexpected token: %u", InfoTok);
+     	}
 
-	}
-
-	/* Directive is followed by a semicolon */
-	InfoConsumeSemi ();
+     	/* Directive is followed by a semicolon */
+     	InfoConsumeSemi ();
     }
 
-    /* Did we get the necessary data */
-    if (Name == 0) {
-	InfoError ("Label name is missing");
+    /* Did we get the necessary data, and is it correct? */
+    if (Name == 0 || Name[0] == '\0') {
+       	InfoError ("Segment name is missing");
     }
-    if (Value < 0) {
-	InfoError ("Label value is missing");
+    if (End < 0) {
+        InfoError ("End address is missing");
     }
-    if (Size < 0) {
-	/* Use default */
-	Size = 1;
+    if (Start < 0) {
+	InfoError ("Start address is missing");
     }
-    if (Value + Size > 0x10000) {
-	InfoError ("Invalid size (address out of range)");
+    if (Start == End) {
+        InfoError ("Segment is empty");
     }
-    if (HaveLabel ((unsigned) Value)) {
-	InfoError ("Label for address $%04lX already defined", Value);
+    if (Start > End) {
+        InfoError ("Start address of segment is greater than end address");
     }
 
-    /* Define the label(s) */
-    AddExtLabelRange ((unsigned) Value, Name, Size);
+    /* Check that segments do not overlap */
+    if (SegmentDefined ((unsigned) Start, (unsigned) End)) {
+        InfoError ("Segments cannot overlap");
+    }
+
+    /* Remember the segment data */
+    AddAbsSegment ((unsigned) Start, (unsigned) End, Name);
 
     /* Delete the dynamically allocated memory for Name */
     xfree (Name);
@@ -440,9 +754,11 @@ static void InfoParse (void)
 /* Parse the config file */
 {
     static const IdentTok Globals[] = {
-     	{   "GLOBAL", 	INFOTOK_GLOBAL	},
-     	{   "RANGE",	INFOTOK_RANGE	},
-	{   "LABEL",	INFOTOK_LABEL	},
+        {   "ASMINC",   INFOTOK_ASMINC  },
+     	{   "GLOBAL", 	INFOTOK_GLOBAL  },
+	{   "LABEL",	INFOTOK_LABEL 	},
+     	{   "RANGE",	INFOTOK_RANGE 	},
+        {   "SEGMENT",  INFOTOK_SEGMENT },
     };
 
     while (InfoTok != INFOTOK_EOF) {
@@ -453,18 +769,28 @@ static void InfoParse (void)
 	/* Check the token */
 	switch (InfoTok) {
 
+            case INFOTOK_ASMINC:
+                AsmIncSection ();
+                break;
+
 	    case INFOTOK_GLOBAL:
 		GlobalSection ();
 		break;
 
-	    case INFOTOK_RANGE:
-		RangeSection ();
-		break;
-
 	    case INFOTOK_LABEL:
 		LabelSection ();
 		break;
 
+	    case INFOTOK_RANGE:
+		RangeSection ();
+		break;
+
+            case INFOTOK_SEGMENT:
+                SegmentSection ();
+                break;
+
+            default:
+                Internal ("Unexpected token: %u", InfoTok);
 	}
 
 	/* Semicolon expected */
