@@ -28,17 +28,17 @@
 #include "grc.h"
 
 /* common stuff */
-//#include "cmdline.h"
 #include "fname.h"
 #include "abend.h"
 #include "chartype.h"
+#include "xmalloc.h"
 
 void VLIRLinker(int argc, char *argv[]) {
 FILE *outCVT, *input;
 unsigned char *buffer;
 unsigned char vlirtabt[127];
 unsigned char vlirtabs[127];
-int i,j,lastarg;         
+int i,j,lastarg;
 unsigned l;
 int bytes;
 int blocks,rest;
@@ -61,7 +61,7 @@ int blocks,rest;
 	if (input==NULL)
 	    AbEnd("can't open input:%s\n",strerror(errno));
 
-	buffer = malloc(THIS_BUFFER_IS_SOOO_HUGE);
+	buffer = xmalloc(THIS_BUFFER_IS_SOOO_HUGE);
 	memset(buffer,0,THIS_BUFFER_IS_SOOO_HUGE);
 
 	bytes = fread(buffer,1,1024,input);
@@ -213,15 +213,15 @@ void openVFile (void) {
 }
 
 void printUsage (void) {
-    fprintf(stderr, "Usage: %s [options] file\n"
-	    "Options:\n"
-	    "\t-h, -?\t\tthis help\n"
-	    "\t-f\t\tforce writting files\n"
-	    "\t-o name\t\tname C output file\n"
-	    "\t-s name\t\tname asm output file\n"
-	    "\t-l name\t\tname ld65 config output file (for vlir)\n"
-	    "Or as VLIR linker: %s -vlir output.cvt header [vlir0] ... [blank] ... [vlir_n]\n",
-	    ProgName,ProgName);
+    printf("Usage: %s [options] file\n"
+           "Options:\n"
+           "\t-h, -?\t\tthis help\n"
+           "\t-f\t\tforce writting files\n"
+           "\t-o name\t\tname C output file\n"
+           "\t-s name\t\tname asm output file\n"
+           "\t-l name\t\tname ld65 config output file (for vlir)\n"
+           "Or as VLIR linker: %s -vlir output.cvt header [vlir0] ... [blank] ... [vlir_n]\n",
+           ProgName,ProgName);
 }
 
 int findToken (const char **tokenTbl, const char *token) {
@@ -261,7 +261,6 @@ int a;
 	}
 }
 
-//char *bintos(unsigned char a, char *out) {
 char *bintos(unsigned char a, char out[7]) {
 int i=0;
     for (;i<8;i++) {
@@ -298,8 +297,8 @@ struct menuitem *curItem, *newItem;
 
     if (strcmp(nextWord(),"{")!=0) {
 	AbEnd ("menu '%s' description has no opening bracket!\n", myMenu.name);
-	};
-    curItem=malloc(sizeof(struct menuitem));
+	}
+    curItem = xmalloc(sizeof(struct menuitem));
     myMenu.item=curItem;
     do {
 	token = nextWord();
@@ -311,13 +310,13 @@ struct menuitem *curItem, *newItem;
 		strcat (namebuff, " ");
 		strcat (namebuff, token);
 		} while (token[strlen(token)-1]!='"');
-	    token = malloc(strlen(namebuff));
+	    token = xmalloc(strlen(namebuff));
 	    strcpy (token, namebuff);
 	}
 	curItem->name=token;
 	curItem->type=nextWord();
 	curItem->target=nextWord();
-	newItem=malloc(sizeof(struct menuitem));
+	newItem=xmalloc(sizeof(struct menuitem));
 	curItem->next=newItem;
 	curItem=newItem;
 	item++;
@@ -392,13 +391,11 @@ int a, b;
 
     a = findToken (hdrFTypes, token);
 
-    if (a>1)
-	   AbEnd("filetype '%s' is not supported yet\n", token);
-
     switch (a) {
-	case 0: myHead.geostype = 6; break;
-	case 1: myHead.geostype = 14; break;
-	}
+    	case 0:  myHead.geostype = 6; break;
+    	case 1:  myHead.geostype = 14; break;
+        default: AbEnd("filetype '%s' is not supported yet\n", token);
+    	}
 
     myHead.dosname = nextPhrase();
     nextPhrase();
@@ -617,13 +614,13 @@ int a, prevchar=-1, i=0, bracket=0, quote=1;
     while (1) {
 	a = getc(F);
 	if ((a=='\n')||(a=='\015')) a = ' ';
-	if (a==',') a = ' ';
+	if (a==',' && quote) a = ' ';
 	if (a=='\042') quote=!quote;
 	if (quote) {
 	    if ((a=='{')||(a=='(')) bracket++;
 	    if ((a=='}')||(a==')')) bracket--;
 	}
-	if (a==EOF) { tbl[i]='\0'; realloc(tbl, i+1); break; };
+	if (a==EOF) { tbl[i]='\0'; xrealloc(tbl, i+1); break; };
 	if (IsSpace(a)) {
 	    if ((prevchar!=' ') && (prevchar!=-1)) { tbl[i++]=' '; prevchar=' '; }
 	} else {
@@ -651,7 +648,7 @@ int vlir=0;	/* number of processed VLIR sections */
     if ((F = fopen (filename,"r"))==0)
 		AbEnd("can't open file %s for reading: %s\n",filename,strerror (errno));
 
-    str=filterInput(F, malloc(BLOODY_BIG_BUFFER));
+    str=filterInput(F, xmalloc(BLOODY_BIG_BUFFER));
 
     token = strtok (str," ");
 
