@@ -6,8 +6,8 @@
 /*                                                                           */
 /*                                                                           */
 /*                                                                           */
-/* (C) 1998-2003 Ullrich von Bassewitz                                       */
-/*               Rmerstrasse 52                                             */
+/* (C) 1998-2008 Ullrich von Bassewitz                                       */
+/*               Roemerstrasse 52                                            */
 /*               D-70794 Filderstadt                                         */
 /* EMail:        uz@cc65.org                                                 */
 /*                                                                           */
@@ -48,12 +48,14 @@
 /* ld65 */
 #include "bin.h"
 #include "binfmt.h"
+#include "cfgexpr.h"
 #include "condes.h"
 #include "config.h"
 #include "error.h"
 #include "exports.h"
 #include "global.h"
 #include "o65.h"
+#include "objdata.h"
 #include "scanner.h"
 #include "spool.h"
 
@@ -65,6 +67,19 @@
 
 
 
+/* Remember which sections we had encountered */
+static enum {
+    SE_NONE     = 0x0000,
+    SE_MEMORY   = 0x0001,
+    SE_SEGMENTS = 0x0002,
+    SE_FEATURES = 0x0004,
+    SE_FILES    = 0x0008,
+    SE_FORMATS  = 0x0010,
+    SE_SYMBOLS  = 0x0020
+} SectionsEncountered = SE_NONE;
+
+
+
 /* File list */
 static File*	    	FileList; 	/* Single linked list */
 static unsigned	    	FileCount;  	/* Number of entries in the list */
@@ -96,10 +111,11 @@ unsigned	       	SegDescCount;	/* Number of entries in list */
 #define SA_LOAD		0x0002
 #define SA_RUN		0x0004
 #define SA_ALIGN	0x0008
-#define SA_DEFINE	0x0010
-#define SA_OFFSET	0x0020
-#define SA_START	0x0040
-#define SA_OPTIONAL     0x0080
+#define SA_ALIGN_LOAD   0x0010
+#define SA_DEFINE	0x0020
+#define SA_OFFSET	0x0040
+#define SA_START	0x0080
+#define SA_OPTIONAL     0x0100
 
 
 
@@ -187,7 +203,7 @@ static Memory* CfgFindMemory (unsigned Name)
 static Memory* CfgGetMemory (unsigned Name)
 /* Find the memory are with the given name. Print an error on an invalid name */
 {
-    Memory* M = CfgFindMemory (Name);
+    Memory* M = CfgFindMemory (Name);   
     if (M == 0) {
  	CfgError ("Invalid memory area `%s'", GetString (Name));
     }
@@ -392,11 +408,11 @@ static void ParseMemory (void)
 {
     static const IdentTok Attributes [] = {
        	{   "START",  	CFGTOK_START    },
-	{   "SIZE",	CFGTOK_SIZE     },
+	{   "SIZE", 	CFGTOK_SIZE     },
         {   "TYPE",     CFGTOK_TYPE     },
         {   "FILE",     CFGTOK_FILE     },
         {   "DEFINE",   CFGTOK_DEFINE   },
-  	{   "FILL",	CFGTOK_FILL     },
+  	{   "FILL", 	CFGTOK_FILL     },
        	{   "FILLVAL", 	CFGTOK_FILLVAL  },
     };
     static const IdentTok Types [] = {
@@ -430,22 +446,21 @@ static void ParseMemory (void)
 
 		case CFGTOK_START:
 		    FlagAttr (&M->Attr, MA_START, "START");
-		    CfgAssureInt ();
-		    M->Start = CfgIVal;
+                    M->Start = CfgIntExpr ();
 		    break;
 
 	      	case CFGTOK_SIZE:
 	     	    FlagAttr (&M->Attr, MA_SIZE, "SIZE");
-		    CfgAssureInt ();
-	      	    M->Size = CfgIVal;
+	      	    M->Size = CfgIntExpr ();
 		    break;
 
 		case CFGTOK_TYPE:
 		    FlagAttr (&M->Attr, MA_TYPE, "TYPE");
       		    CfgSpecialToken (Types, ENTRY_COUNT (Types), "Type");
 		    if (CfgTok == CFGTOK_RO) {
-			M->Flags |= MF_RO;
+		    	M->Flags |= MF_RO;
 		    }
+                    CfgNextTok ();
 		    break;
 
 	        case CFGTOK_FILE:
@@ -453,6 +468,7 @@ static void ParseMemory (void)
 		    CfgAssureStr ();
        	       	    /* Get the file entry and insert the memory area */
 	    	    FileInsert (GetFile (GetStringId (CfgSVal)), M);
+                    CfgNextTok ();
 		    break;
 
 	        case CFGTOK_DEFINE:
@@ -462,6 +478,7 @@ static void ParseMemory (void)
 		    if (CfgTok == CFGTOK_TRUE) {
 	  	    	M->Flags |= MF_DEFINE;
 		    }
+                    CfgNextTok ();
 		    break;
 
 	        case CFGTOK_FILL:
@@ -471,13 +488,12 @@ static void ParseMemory (void)
 		    if (CfgTok == CFGTOK_TRUE) {
 	  	    	M->Flags |= MF_FILL;
 		    }
+                    CfgNextTok ();
 		    break;
 
 	      	case CFGTOK_FILLVAL:
 		    FlagAttr (&M->Attr, MA_FILLVAL, "FILLVAL");
-	     	    CfgAssureInt ();
-		    CfgRangeCheck (0, 0xFF);
-	      	    M->FillVal = (unsigned char) CfgIVal;
+	      	    M->FillVal = (unsigned char) CfgCheckedIntExpr (0, 0xFF);
 		    break;
 
 	     	default:
@@ -485,8 +501,7 @@ static void ParseMemory (void)
 
 	    }
 
-	    /* Skip the attribute value and an optional comma */
-      	    CfgNextTok ();
+	    /* Skip an optional comma */
 	    CfgOptionalComma ();
 	}
 
@@ -504,6 +519,9 @@ static void ParseMemory (void)
 	    FileInsert (GetFile (GetStringId (OutputName)), M);
 	}
     }
+
+    /* Remember we had this section */
+    SectionsEncountered |= SE_MEMORY;
 }
 
 
@@ -515,12 +533,17 @@ static void ParseFiles (void)
        	{   "FORMAT",  	CFGTOK_FORMAT   },
     };
     static const IdentTok Formats [] = {
-       	{   "O65",     	CFGTOK_O65	     },
+       	{   "O65",     	CFGTOK_O65  	     },
        	{   "BIN",     	CFGTOK_BIN      },
        	{   "BINARY",   CFGTOK_BIN      },
     };
 
 
+    /* The MEMORY section must preceed the FILES section */
+    if ((SectionsEncountered & SE_MEMORY) == 0) {
+        CfgError ("MEMORY must precede FILES");
+    }
+
     /* Parse all files */
     while (CfgTok != CFGTOK_RCURLY) {
 
@@ -532,7 +555,7 @@ static void ParseFiles (void)
 	/* Search for the file, it must exist */
        	F = FindFile (GetStringId (CfgSVal));
 	if (F == 0) {
-	    CfgError ("No such file: `%s'", CfgSVal);
+       	    CfgError ("File `%s' not found in MEMORY section", CfgSVal);
 	}
 
 	/* Skip the token and the following colon */
@@ -590,6 +613,9 @@ static void ParseFiles (void)
 	CfgConsumeSemi ();
 
     }
+
+    /* Remember we had this section */
+    SectionsEncountered |= SE_FILES;
 }
 
 
@@ -598,24 +624,30 @@ static void ParseSegments (void)
 /* Parse a SEGMENTS section */
 {
     static const IdentTok Attributes [] = {
-        {   "ALIGN",    CFGTOK_ALIGN    },
-        {   "DEFINE",   CFGTOK_DEFINE   },
-       	{   "LOAD",    	CFGTOK_LOAD     },
-	{   "OFFSET",  	CFGTOK_OFFSET   },
-        {   "OPTIONAL", CFGTOK_OPTIONAL },
-	{   "RUN",     	CFGTOK_RUN      },
-	{   "START",   	CFGTOK_START    },
-        {   "TYPE",     CFGTOK_TYPE     },
+        {   "ALIGN",            CFGTOK_ALIGN            },
+        {   "ALIGN_LOAD",       CFGTOK_ALIGN_LOAD       },
+        {   "DEFINE",           CFGTOK_DEFINE           },
+       	{   "LOAD",    	        CFGTOK_LOAD             },
+	{   "OFFSET",  	        CFGTOK_OFFSET           },
+        {   "OPTIONAL",         CFGTOK_OPTIONAL         },
+	{   "RUN",     	        CFGTOK_RUN              },
+	{   "START",   	        CFGTOK_START            },
+        {   "TYPE",             CFGTOK_TYPE             },
     };
     static const IdentTok Types [] = {
-       	{   "RO",      	CFGTOK_RO       },
-       	{   "RW",      	CFGTOK_RW       },
-       	{   "BSS",     	CFGTOK_BSS      },
-	{   "ZP",      	CFGTOK_ZP	},
-        {   "WPROT",    CFGTOK_RO       },      /* ### OBSOLETE */
+       	{   "RO",      	        CFGTOK_RO               },
+       	{   "RW",      	        CFGTOK_RW               },
+       	{   "BSS",     	        CFGTOK_BSS              },
+	{   "ZP",      	        CFGTOK_ZP	        },
     };
 
     unsigned Count;
+    long     Val;
+
+    /* The MEMORY section must preceed the SEGMENTS section */
+    if ((SectionsEncountered & SE_MEMORY) == 0) {
+        CfgError ("MEMORY must precede SEGMENTS");
+    }
 
     while (CfgTok == CFGTOK_IDENT) {
 
@@ -644,16 +676,25 @@ static void ParseSegments (void)
 	    switch (AttrTok) {
 
 	        case CFGTOK_ALIGN:
-	    	    CfgAssureInt ();
 	    	    FlagAttr (&S->Attr, SA_ALIGN, "ALIGN");
-	    	    CfgRangeCheck (1, 0x10000);
-	    	    S->Align = BitFind (CfgIVal);
-	    	    if ((0x01UL << S->Align) != CfgIVal) {
+	    	    Val = CfgCheckedIntExpr (1, 0x10000);
+	    	    S->Align = BitFind (Val);
+	    	    if ((0x01L << S->Align) != Val) {
 	    	     	CfgError ("Alignment must be a power of 2");
 	    	    }
 	    	    S->Flags |= SF_ALIGN;
 	    	    break;
 
+                case CFGTOK_ALIGN_LOAD:
+	    	    FlagAttr (&S->Attr, SA_ALIGN_LOAD, "ALIGN_LOAD");
+	    	    Val = CfgCheckedIntExpr (1, 0x10000);
+       	       	    S->AlignLoad = BitFind (Val);
+	    	    if ((0x01L << S->AlignLoad) != Val) {
+	    	     	CfgError ("Alignment must be a power of 2");
+	    	    }
+	    	    S->Flags |= SF_ALIGN_LOAD;
+	    	    break;
+
 	        case CFGTOK_DEFINE:
  	    	    FlagAttr (&S->Attr, SA_DEFINE, "DEFINE");
 	    	    /* Map the token to a boolean */
@@ -661,18 +702,18 @@ static void ParseSegments (void)
 	    	    if (CfgTok == CFGTOK_TRUE) {
 	    	     	S->Flags |= SF_DEFINE;
 	    	    }
+                    CfgNextTok ();
 	    	    break;
 
 	    	case CFGTOK_LOAD:
 	      	    FlagAttr (&S->Attr, SA_LOAD, "LOAD");
 	    	    S->Load = CfgGetMemory (GetStringId (CfgSVal));
+                    CfgNextTok ();
 	    	    break;
 
 	        case CFGTOK_OFFSET:
-	    	    CfgAssureInt ();
 	    	    FlagAttr (&S->Attr, SA_OFFSET, "OFFSET");
-	    	    CfgRangeCheck (1, 0x1000000);
-	    	    S->Addr   = CfgIVal;
+	    	    S->Addr   = CfgCheckedIntExpr (1, 0x1000000);
 	    	    S->Flags |= SF_OFFSET;
 	    	    break;
 
@@ -682,18 +723,18 @@ static void ParseSegments (void)
 		    if (CfgTok == CFGTOK_TRUE) {
 	  	    	S->Flags |= SF_OPTIONAL;
 		    }
+                    CfgNextTok ();
 	    	    break;
 
 	    	case CFGTOK_RUN:
       	    	    FlagAttr (&S->Attr, SA_RUN, "RUN");
  	    	    S->Run = CfgGetMemory (GetStringId (CfgSVal));
+                    CfgNextTok ();
 	    	    break;
 
 	        case CFGTOK_START:
-      	    	    CfgAssureInt ();
 	    	    FlagAttr (&S->Attr, SA_START, "START");
-	    	    CfgRangeCheck (1, 0x1000000);
-	    	    S->Addr   = CfgIVal;
+	    	    S->Addr   = CfgCheckedIntExpr (1, 0x1000000);
 	    	    S->Flags |= SF_START;
 	    	    break;
 
@@ -705,8 +746,9 @@ static void ParseSegments (void)
 	    		case CFGTOK_RW:	   /* Default */		    break;
 	    	     	case CFGTOK_BSS:   S->Flags |= SF_BSS;              break;
 	    	     	case CFGTOK_ZP:	   S->Flags |= (SF_BSS | SF_ZP);    break;
-	    	     	default:	   Internal ("Unexpected token: %d", CfgTok);
+	    	     	default:      	   Internal ("Unexpected token: %d", CfgTok);
 	    	    }
+                    CfgNextTok ();
 	    	    break;
 
 	    	default:
@@ -714,8 +756,7 @@ static void ParseSegments (void)
 
 	    }
 
-	    /* Skip the attribute value and an optional comma */
-	    CfgNextTok ();
+	    /* Skip an optional comma */
 	    CfgOptionalComma ();
 	}
 
@@ -726,13 +767,6 @@ static void ParseSegments (void)
 	if ((S->Attr & SA_RUN) == 0) {
 	    S->Attr |= SA_RUN;
 	    S->Run = S->Load;
-	} else {
-	    /* Both attributes given */
-	    S->Flags |= SF_LOAD_AND_RUN;
-	}
-	if ((S->Attr & SA_ALIGN) == 0) {
-	    S->Attr |= SA_ALIGN;
-	    S->Align = 0;
 	}
 
 	/* If the segment is marked as BSS style, and if the segment exists
@@ -744,10 +778,21 @@ static void ParseSegments (void)
 	    	     CfgGetName (), CfgErrorLine);
 	}
 
+        /* An attribute of ALIGN_LOAD doesn't make sense if there are no
+         * separate run and load memory areas.
+         */
+        if ((S->Flags & SF_ALIGN_LOAD) != 0 && (S->Load == S->Run)) {
+       	    Warning ("%s(%u): ALIGN_LOAD attribute specified, but no separate "
+                     "LOAD and RUN memory areas assigned",
+                     CfgGetName (), CfgErrorLine);
+            /* Remove the flag */
+            S->Flags &= ~SF_ALIGN_LOAD;
+        }
+
         /* If the segment is marked as BSS style, it may not have separate
          * load and run memory areas, because it's is never written to disk.
          */
-        if ((S->Flags & SF_BSS) != 0 && (S->Flags & SF_LOAD_AND_RUN) != 0) {
+        if ((S->Flags & SF_BSS) != 0 && (S->Load != S->Run)) {
        	    Warning ("%s(%u): Segment with type `bss' has both LOAD and RUN "
                      "memory areas assigned", CfgGetName (), CfgErrorLine);
         }
@@ -778,8 +823,8 @@ static void ParseSegments (void)
 	    SegDescInsert (S);
       	    /* Insert the segment into the memory area list */
       	    MemoryInsert (S->Run, S);
-      	    if ((S->Flags & SF_LOAD_AND_RUN) != 0) {
-      	    	/* We have a separate RUN area given */
+      	    if (S->Load != S->Run) {
+      	    	/* We have separate RUN and LOAD areas */
       	    	MemoryInsert (S->Load, S);
       	    }
       	} else {
@@ -794,6 +839,9 @@ static void ParseSegments (void)
 	/* Skip the semicolon */
 	CfgConsumeSemi ();
     }
+
+    /* Remember we had this section */
+    SectionsEncountered |= SE_SEGMENTS;
 }
 
 
@@ -817,6 +865,7 @@ static void ParseO65 (void)
        	{   "LUNIX",   	CFGTOK_LUNIX     	},
        	{   "OSA65",   	CFGTOK_OSA65    	},
         {   "CC65",     CFGTOK_CC65             },
+        {   "OPENCBM",  CFGTOK_OPENCBM          },
     };
 
     /* Bitmask to remember the attributes we got already */
@@ -864,7 +913,7 @@ static void ParseO65 (void)
 		    CfgError ("Exported symbol `%s' cannot be an import", CfgSVal);
 		}
       		/* Check if we have this symbol defined already. The entry
-      		 * routine will check this also, but we get a more verbose
+      	     	 * routine will check this also, but we get a more verbose
       		 * error message when checking it here.
       		 */
       	 	if (O65GetExport (O65FmtDesc, CfgSValId) != 0) {
@@ -872,6 +921,8 @@ static void ParseO65 (void)
       	 	}
 		/* Insert the symbol into the table */
 	  	O65SetExport (O65FmtDesc, CfgSValId);
+                /* Eat the identifier token */
+                CfgNextTok ();
 	    	break;
 
 	    case CFGTOK_IMPORT:
@@ -894,6 +945,8 @@ static void ParseO65 (void)
       	    	}
 	    	/* Insert the symbol into the table */
 	    	O65SetImport (O65FmtDesc, CfgSValId);
+                /* Eat the identifier token */
+                CfgNextTok ();
 	    	break;
 
 	    case CFGTOK_TYPE:
@@ -914,37 +967,44 @@ static void ParseO65 (void)
 	    	    default:
 	    	     	CfgError ("Unexpected type token");
 	    	}
-	    	break;
+                /* Eat the attribute token */
+                CfgNextTok ();
+	     	break;
 
 	    case CFGTOK_OS:
-		/* Cannot use this attribute twice */
-	  	FlagAttr (&AttrFlags, atOS, "OS");
-		/* Get the operating system */
-	    	CfgSpecialToken (OperatingSystems, ENTRY_COUNT (OperatingSystems), "OS type");
-		switch (CfgTok) {
-		    case CFGTOK_LUNIX:  OS = O65OS_LUNIX;       break;
-		    case CFGTOK_OSA65:  OS = O65OS_OSA65;       break;
-		    case CFGTOK_CC65:   OS = O65OS_CC65;        break;
-		    default:            CfgError ("Unexpected OS token");
-		}
-		break;
+	     	/* Cannot use this attribute twice */
+	     	FlagAttr (&AttrFlags, atOS, "OS");
+	     	/* Get the operating system. It may be specified as name or
+                 * as a number in the range 1..255.
+                 */
+		if (CfgTok == CFGTOK_INTCON) {
+		    CfgRangeCheck (O65OS_MIN, O65OS_MAX);
+		    OS = (unsigned) CfgIVal;
+		} else {
+                    CfgSpecialToken (OperatingSystems, ENTRY_COUNT (OperatingSystems), "OS type");
+                    switch (CfgTok) {
+                        case CFGTOK_LUNIX:    OS = O65OS_LUNIX;     break;
+                        case CFGTOK_OSA65:    OS = O65OS_OSA65;     break;
+                        case CFGTOK_CC65:     OS = O65OS_CC65;      break;
+                        case CFGTOK_OPENCBM:  OS = O65OS_OPENCBM;   break;
+                        default:              CfgError ("Unexpected OS token");
+                    }
+                }
+                CfgNextTok ();
+	     	break;
 
             case CFGTOK_ID:
                 /* Cannot have this attribute twice */
                 FlagAttr (&AttrFlags, atID, "ID");
                 /* We're expecting a number in the 0..$FFFF range*/
-                CfgAssureInt ();
-                CfgRangeCheck (0, 0xFFFF);
-                ModuleId = (unsigned) CfgIVal;
+                ModuleId = (unsigned) CfgCheckedIntExpr (0, 0xFFFF);
                 break;
 
             case CFGTOK_VERSION:
                 /* Cannot have this attribute twice */
                 FlagAttr (&AttrFlags, atVersion, "VERSION");
                 /* We're expecting a number in byte range */
-                CfgAssureInt ();
-                CfgRangeCheck (0, 0xFF);
-                Version = (unsigned) CfgIVal;
+                Version = (unsigned) CfgCheckedIntExpr (0, 0xFF);
                 break;
 
 	    default:
@@ -952,8 +1012,7 @@ static void ParseO65 (void)
 
 	}
 
-	/* Skip the attribute value and an optional comma */
-	CfgNextTok ();
+	/* Skip an optional comma */
 	CfgOptionalComma ();
     }
 
@@ -962,8 +1021,8 @@ static void ParseO65 (void)
 
     /* Check for attributes that may not be combined */
     if (OS == O65OS_CC65) {
-        if ((AttrFlags & (atImport | atExport)) != 0) {
-            CfgError ("OS type CC65 may not have imports or exports");
+        if ((AttrFlags & (atImport | atExport)) != 0 && ModuleId < 0x8000) {
+            CfgError ("OS type CC65 may not have imports or exports for ids < $8000");
         }
     } else {
         if (AttrFlags & atID) {
@@ -1015,6 +1074,10 @@ static void ParseFormats (void)
 	/* Skip the semicolon */
 	CfgConsumeSemi ();
     }
+
+
+    /* Remember we had this section */
+    SectionsEncountered |= SE_FORMATS;
 }
 
 
@@ -1033,6 +1096,7 @@ static void ParseConDes (void)
     static const IdentTok Types [] = {
        	{   "CONSTRUCTOR",	CFGTOK_CONSTRUCTOR	},
 	{   "DESTRUCTOR",      	CFGTOK_DESTRUCTOR	},
+        {   "INTERRUPTOR",      CFGTOK_INTERRUPTOR      },
     };
 
     static const IdentTok Orders [] = {
@@ -1113,6 +1177,7 @@ static void ParseConDes (void)
 		    switch (CfgTok) {
 		     	case CFGTOK_CONSTRUCTOR: Type = CD_TYPE_CON;	break;
 		     	case CFGTOK_DESTRUCTOR:	 Type = CD_TYPE_DES;	break;
+                        case CFGTOK_INTERRUPTOR: Type = CD_TYPE_INT;    break;
 	     	     	default: FAIL ("Unexpected type token");
 		    }
 		}
@@ -1204,11 +1269,8 @@ static void ParseStartAddress (void)
 	    case CFGTOK_DEFAULT:
 	  	/* Don't allow this twice */
 		FlagAttr (&AttrFlags, atDefault, "DEFAULT");
-	      	/* We expect a number */
-		CfgAssureInt ();
-                CfgRangeCheck (0, 0xFFFFFF);
-		/* Remember the value for later */
-                DefStartAddr = CfgIVal;
+	      	/* We expect a numeric expression */
+                DefStartAddr = CfgCheckedIntExpr (0, 0xFFFFFF);
 	    	break;
 
 	    default:
@@ -1216,9 +1278,6 @@ static void ParseStartAddress (void)
 
 	}
 
-       	/* Skip the attribute value */
-	CfgNextTok ();
-
 	/* Semicolon ends the ConDes decl, otherwise accept an optional comma */
 	if (CfgTok == CFGTOK_SEMI) {
 	    break;
@@ -1263,8 +1322,8 @@ static void ParseFeatures (void)
     	switch (FeatureTok) {
 
     	    case CFGTOK_CONDES:
-    		ParseConDes ();
-    		break;
+    	 	ParseConDes ();
+    	 	break;
 
             case CFGTOK_STARTADDRESS:
                 ParseStartAddress ();
@@ -1272,12 +1331,15 @@ static void ParseFeatures (void)
 
 
     	    default:
-    		Error ("Unexpected feature token");
+       	       	FAIL ("Unexpected feature token");
     	}
 
     	/* Skip the semicolon */
     	CfgConsumeSemi ();
     }
+
+    /* Remember we had this section */
+    SectionsEncountered |= SE_FEATURES;
 }
 
 
@@ -1285,28 +1347,125 @@ static void ParseFeatures (void)
 static void ParseSymbols (void)
 /* Parse a symbols section */
 {
+    static const IdentTok Attributes[] = {
+       	{   "VALUE",   	CFGTOK_VALUE    },
+        {   "WEAK",     CFGTOK_WEAK     },
+    };
+
     while (CfgTok == CFGTOK_IDENT) {
 
-	long Val;
+	long Val = 0L;
+        int  Weak = 0;
+        Export* E;
 
 	/* Remember the name */
 	unsigned Name = GetStringId (CfgSVal);
 	CfgNextTok ();
 
-	/* Allow an optional assignment */
-	CfgOptionalAssign ();
+        /* Support both, old and new syntax here. New syntax is a colon
+         * followed by an attribute list, old syntax is an optional equal
+         * sign plus a value.
+         */
+        if (CfgTok != CFGTOK_COLON) {
 
-	/* Make sure the next token is an integer, read and skip it */
-	CfgAssureInt ();
-	Val = CfgIVal;
-	CfgNextTok ();
+            /* Old syntax */
 
-	/* Generate an export with the given value */
-	CreateConstExport (Name, Val);
+            /* Allow an optional assignment */
+            CfgOptionalAssign ();
 
-	/* Skip the semicolon */
-	CfgConsumeSemi ();
+            /* Make sure the next token is an integer expression, read and
+             * skip it.
+             */
+            Val = CfgIntExpr ();
+
+        } else {
+
+            /* Bitmask to remember the attributes we got already */
+            enum {
+                atNone	     	= 0x0000,
+                atValue         = 0x0001,
+                atWeak          = 0x0002
+            };
+            unsigned AttrFlags = atNone;
+
+
+            /* New syntax - skip the colon */
+            CfgNextTok ();
+
+            /* Parse the attributes */
+            while (1) {
+
+                /* Map the identifier to a token */
+                cfgtok_t AttrTok;
+                CfgSpecialToken (Attributes, ENTRY_COUNT (Attributes), "Attribute");
+                AttrTok = CfgTok;
+
+                /* Skip the attribute name */
+                CfgNextTok ();
+
+                /* An optional assignment follows */
+                CfgOptionalAssign ();
+
+                /* Check which attribute was given */
+                switch (AttrTok) {
+
+                    case CFGTOK_VALUE:
+                        /* Don't allow this twice */
+                        FlagAttr (&AttrFlags, atValue, "VALUE");
+                        /* We expect a numeric expression */
+                        Val = CfgIntExpr ();
+                        break;
+
+                    case CFGTOK_WEAK:
+                        /* Don't allow this twice */
+                        FlagAttr (&AttrFlags, atWeak, "WEAK");
+                        CfgBoolToken ();
+                        Weak = (CfgTok == CFGTOK_TRUE);
+                        CfgNextTok ();
+                        break;
+
+                    default:
+                        FAIL ("Unexpected attribute token");
+
+                }
+
+                /* Semicolon ends the decl, otherwise accept an optional comma */
+                if (CfgTok == CFGTOK_SEMI) {
+                    break;
+                } else if (CfgTok == CFGTOK_COMMA) {
+                    CfgNextTok ();
+                }
+            }
+
+            /* Check if we have all mandatory attributes */
+            AttrCheck (AttrFlags, atValue, "VALUE");
+
+            /* Weak is optional, the default are non weak symbols */
+            if ((AttrFlags & atWeak) == 0) {
+                Weak = 0;
+            }
+
+        }
+
+        /* Check if the symbol is already defined */
+        if ((E = FindExport (Name)) != 0 && !IsUnresolvedExport (E)) {
+            /* If the symbol is not marked as weak, this is an error.
+             * Otherwise ignore the symbol from the config.
+             */
+            if (!Weak) {
+                CfgError ("Symbol `%s' is already defined", GetString (Name));
+            }
+        } else {
+            /* The symbol is undefined, generate an export */
+            CreateConstExport (Name, Val);
+        }
+
+    	/* Skip the semicolon */
+    	CfgConsumeSemi ();
     }
+
+    /* Remember we had this section */
+    SectionsEncountered |= SE_SYMBOLS;
 }
 
 
@@ -1398,13 +1557,14 @@ void CfgRead (void)
 static void CreateRunDefines (SegDesc* S, unsigned long SegAddr)
 /* Create the defines for a RUN segment */
 {
-    char Buf [256];
+    StrBuf Buf = STATIC_STRBUF_INITIALIZER;
 
-    xsprintf (Buf, sizeof (Buf), "__%s_RUN__", GetString (S->Name));
-    CreateMemoryExport (GetStringId (Buf), S->Run, SegAddr - S->Run->Start);
-    xsprintf (Buf, sizeof (Buf), "__%s_SIZE__", GetString (S->Name));
-    CreateConstExport (GetStringId (Buf), S->Seg->Size);
+    SB_Printf (&Buf, "__%s_RUN__", GetString (S->Name));
+    CreateMemoryExport (GetStrBufId (&Buf), S->Run, SegAddr - S->Run->Start);
+    SB_Printf (&Buf, "__%s_SIZE__", GetString (S->Name));
+    CreateConstExport (GetStrBufId (&Buf), S->Seg->Size);
     S->Flags |= SF_RUN_DEF;
+    SB_Done (&Buf);
 }
 
 
@@ -1412,18 +1572,25 @@ static void CreateRunDefines (SegDesc* S, unsigned long SegAddr)
 static void CreateLoadDefines (SegDesc* S, unsigned long SegAddr)
 /* Create the defines for a LOAD segment */
 {
-    char Buf [256];
+    StrBuf Buf = STATIC_STRBUF_INITIALIZER;
 
-    xsprintf (Buf, sizeof (Buf), "__%s_LOAD__", GetString (S->Name));
-    CreateMemoryExport (GetStringId (Buf), S->Load, SegAddr - S->Load->Start);
+    SB_Printf (&Buf, "__%s_LOAD__", GetString (S->Name));
+    CreateMemoryExport (GetStrBufId (&Buf), S->Load, SegAddr - S->Load->Start);
     S->Flags |= SF_LOAD_DEF;
+    SB_Done (&Buf);
 }
 
 
 
-void CfgAssignSegments (void)
-/* Assign segments, define linker symbols where requested */
+unsigned CfgAssignSegments (void)
+/* Assign segments, define linker symbols where requested. The function will
+ * return the number of memory area overflows (so zero means anything went ok).
+ * In case of overflows, a short mapfile can be generated later, to ease the
+ * task of rearranging segments for the user.
+ */
 {
+    unsigned Overflows = 0;
+
     /* Walk through each of the memory sections. Add up the sizes and check
      * for an overflow of the section. Assign the start addresses of the
      * segments while doing this.
@@ -1446,82 +1613,81 @@ void CfgAssignSegments (void)
      	    /* Get the segment from the node */
      	    SegDesc* S = N->Seg;
 
-     	    /* Handle ALIGN and OFFSET/START */
-     	    if (S->Flags & SF_ALIGN) {
-     	   	/* Align the address */
-     	   	unsigned long Val = (0x01UL << S->Align) - 1;
-     	   	Addr = (Addr + Val) & ~Val;
-     	    } else if (S->Flags & (SF_OFFSET | SF_START)) {
-     	   	/* Give the segment a fixed starting address */
-     	   	unsigned long NewAddr = S->Addr;
-     	   	if (S->Flags & SF_OFFSET) {
-     	   	    /* An offset was given, no address, make an address */
-     	   	    NewAddr += M->Start;
-     	   	}
-       	       	if (Addr > NewAddr) {
-     	   	    /* Offset already too large */
-     	   	    if (S->Flags & SF_OFFSET) {
-     	   	        Error ("Offset too small in `%s', segment `%s'",
-     	   	     	       GetString (M->Name), GetString (S->Name));
-     	   	    } else {
-     	   	     	Error ("Start address too low in `%s', segment `%s'",
-     	   	     	       GetString (M->Name), GetString (S->Name));
-     	   	    }
-     	   	}
-     	   	Addr = NewAddr;
-     	    }
-
-       	    /* If this is the run area, set the start address of this segment,
-             * set the readonly flag in the segment and and remember if the 
-             * segment is in a relocatable file or not.
+            /* Some actions depend on wether this is the load or run memory
+             * area.
              */
-     	    if (S->Run == M) {
-     	        S->Seg->PC = Addr;
+            if (S->Run == M) {
+
+                /* This is the run (and maybe load) memory area. Handle
+                 * alignment and explict start address and offset.
+                 */
+                if (S->Flags & SF_ALIGN) {
+                    /* Align the address */
+                    unsigned long Val = (0x01UL << S->Align) - 1;
+                    Addr = (Addr + Val) & ~Val;
+                } else if (S->Flags & (SF_OFFSET | SF_START)) {
+                    /* Give the segment a fixed starting address */
+                    unsigned long NewAddr = S->Addr;
+                    if (S->Flags & SF_OFFSET) {
+                        /* An offset was given, no address, make an address */
+                        NewAddr += M->Start;
+                    }
+                    if (Addr > NewAddr) {
+                        /* Offset already too large */
+                        if (S->Flags & SF_OFFSET) {
+                            Error ("Offset too small in `%s', segment `%s'",
+                                   GetString (M->Name), GetString (S->Name));
+                        } else {
+                            Error ("Start address too low in `%s', segment `%s'",
+                                   GetString (M->Name), GetString (S->Name));
+                        }
+                    }
+                    Addr = NewAddr;
+                }
+
+                /* Set the start address of this segment, set the readonly flag
+                 * in the segment and and remember if the segment is in a
+                 * relocatable file or not.
+                 */
+                S->Seg->PC = Addr;
                 S->Seg->ReadOnly = (S->Flags & SF_RO) != 0;
                 S->Seg->Relocatable = M->Relocatable;
-     	    }
+
+            } else if (S->Load == M) {
+
+                /* This is the load memory area, *and* run and load are
+                 * different (because of the "else" above). Handle alignment.
+                 */
+                if (S->Flags & SF_ALIGN_LOAD) {
+                    /* Align the address */
+                    unsigned long Val = (0x01UL << S->AlignLoad) - 1;
+                    Addr = (Addr + Val) & ~Val;
+                }
+
+            }
 
      	    /* Increment the fill level of the memory area and check for an
      	     * overflow.
      	     */
      	    M->FillLevel = Addr + S->Seg->Size - M->Start;
-     	    if (M->FillLevel > M->Size) {
-     	     	Error ("Memory area overflow in `%s', segment `%s' (%lu bytes)",
-     	 	       GetString (M->Name), GetString (S->Name),
-                       M->FillLevel - M->Size);
+       	    if (M->FillLevel > M->Size && (M->Flags & MF_OVERFLOW) == 0) {
+                ++Overflows;
+                M->Flags |= MF_OVERFLOW;
+                Warning ("Memory area overflow in `%s', segment `%s' (%lu bytes)",
+                         GetString (M->Name), GetString (S->Name),
+                         M->FillLevel - M->Size);
      	    }
 
      	    /* If requested, define symbols for the start and size of the
      	     * segment.
      	     */
      	    if (S->Flags & SF_DEFINE) {
-		if ((S->Flags & SF_LOAD_AND_RUN) && S->Run == S->Load) {
-		    /* RUN and LOAD given and in one memory area.
-		     * Be careful: We will encounter this code twice, the
-		     * first time when walking the RUN list, second time when
-		     * walking the LOAD list. Be sure to define only the
-		     * relevant symbols on each walk.
-		     */
-		    if (S->Load == M) {
-		       	if ((S->Flags & SF_LOAD_DEF) == 0) {
-		       	    CreateLoadDefines (S, Addr);
-		       	} else {
-		       	    CHECK ((S->Flags & SF_RUN_DEF) == 0);
-		       	    CreateRunDefines (S, Addr);
-		       	}
-		    }
-		} else {
-		    /* RUN and LOAD in different memory areas, or RUN not
-     		     * given, so RUN defaults to LOAD. In the latter case, we
-		     * have only one copy of the segment in the area.
-		     */
-		    if (S->Run == M) {
-		       	CreateRunDefines (S, Addr);
-		    }
-		    if (S->Load == M) {
-		       	CreateLoadDefines (S, Addr);
-		    }
-		}
+                if (S->Run == M && (S->Flags & SF_RUN_DEF) == 0) {
+                    CreateRunDefines (S, Addr);
+                }
+                if (S->Load == M && (S->Flags & SF_LOAD_DEF) == 0) {
+                    CreateLoadDefines (S, Addr);
+                }
      	    }
 
      	    /* Calculate the new address */
@@ -1533,18 +1699,22 @@ void CfgAssignSegments (void)
 
 	/* If requested, define symbols for start and size of the memory area */
 	if (M->Flags & MF_DEFINE) {
-	    char Buf [256];
-	    sprintf (Buf, "__%s_START__", GetString (M->Name));
-	    CreateMemoryExport (GetStringId (Buf), M, 0);
-	    sprintf (Buf, "__%s_SIZE__", GetString (M->Name));
-	    CreateConstExport (GetStringId (Buf), M->Size);
-	    sprintf (Buf, "__%s_LAST__", GetString (M->Name));
-	    CreateConstExport (GetStringId (Buf), M->FillLevel);
+	    StrBuf Buf = STATIC_STRBUF_INITIALIZER;
+	    SB_Printf (&Buf, "__%s_START__", GetString (M->Name));
+	    CreateMemoryExport (GetStrBufId (&Buf), M, 0);
+	    SB_Printf (&Buf, "__%s_SIZE__", GetString (M->Name));
+	    CreateConstExport (GetStrBufId (&Buf), M->Size);
+	    SB_Printf (&Buf, "__%s_LAST__", GetString (M->Name));
+	    CreateMemoryExport (GetStrBufId (&Buf), M, M->FillLevel);
+            SB_Done (&Buf);
 	}
 
 	/* Next memory area */
 	M = M->Next;
     }
+
+    /* Return the number of memory area overflows */
+    return Overflows;
 }
 
 
@@ -1561,7 +1731,7 @@ void CfgWriteTarget (void)
   	if (F->MemList) {
 
   	    /* Is there an output file? */
-  	    if (strlen (GetString (F->Name)) > 0) {
+  	    if (SB_GetLen (GetStrBuf (F->Name)) > 0) {
 
   		/* Assign a proper binary format */
   		if (F->Format == BINFMT_DEFAULT) {
